def clean_data(elems=None, selector=None):
    if selector.get('multiple'):
        data_cleaned = []
        data = elems.extract()
        for i, datum in enumerate(data):
            if datum:
                data_cleaned.append(datum.strip())
        return data_cleaned
    else:
        data = elems.extract_first()
        if data:
            return data.strip()
        return data


def get_selector_element(html_element, selector, ):
    if selector.get('selector_attribute') in ['text']:
        if selector.get('selector_type') == 'css':
            elems = html_element.css("{0}::{1}".format(selector.get('selector'),
                                                       selector.get('selector_attribute')))
            return clean_data(elems=elems, selector=selector)
        else:
            raise NotImplemented("selector_type not equal to css; this is not implemented")
    elif selector.get('selector_attribute') == 'html':
        if selector.get('selector_type') == 'css':
            elems = html_element.css(selector.get('selector'))
            return clean_data(elems=elems, selector=selector)
        else:
            raise NotImplemented("selector_type not equal to css; this is not implemented")
    else:
        if selector.get('selector_type') == 'css':
            elems = html_element.css(selector.get('selector')) \
                .xpath("@{0}".format(selector.get('selector_attribute')))
            return clean_data(elems=elems, selector=selector)
        else:
            raise NotImplemented("selector_type not equal to css; this is not implemented")



from invana_bot.utils.config import validate_config, process_config


def process_pipe(parser_config=None):
    return process_config(parser_config)


def process_pipeline_config(pipeline=None):
    # TODO - validate the config too
    for pipe in pipeline['pipeline']:
        for extractor in pipe.get('data_extractors', []):
            extractor['data_selectors'] = process_pipe(extractor)['data_selectors']

    return pipeline



from invana_bot.settings import MONGODB_DEFAULTS, ELASTICSEARCH_DEFAULTS, \
    FEEDS_CRAWLER_DEFAULTS, WEBSITE_CRAWLER_DEFAULTS, SUPPORTED_DATABASES, SUPPORTED_CRAWLERS
from invana_bot.pipelines.default import WebCrawlerPipeline, WebCrawlerPipelet
from scrapy.crawler import CrawlerRunner, CrawlerProcess
from scrapy.crawler import CrawlerProcess
from invana_bot.spiders.feeds import RSSSpider
from twisted.internet import reactor
import uuid
from datetime import datetime

from scrapy.utils.log import configure_logging


class InvanaCrawlerBase(object):
    """


    """
    settings = {
        'COMPRESSION_ENABLED': False,  # this will save the data in normal text form, otherwise to bytes form.
        'HTTPCACHE_ENABLED': True,
        'TELNETCONSOLE_PORT': None

    }
    runner = None

    def __init__(self,
                 cache_database=None,
                 storage_database=None,
                 cache_database_name=None,
                 storage_database_name=None,

                 cache_database_uri=None,
                 storage_database_uri=None,
                 http_cache_enabled=True,
                 log_level="DEBUG",
                 extra_settings=None,
                 **kwargs):

        self.settings['HTTPCACHE_ENABLED'] = http_cache_enabled
        self.settings['LOG_LEVEL'] = log_level
        if extra_settings:
            self.settings.update(extra_settings)  # over riding or adding extra settings
        self.cache_database_uri = cache_database_uri
        self.storage_database_uri = storage_database_uri

        self.setup_database_settings(cache_database=cache_database,
                                     storage_database=storage_database
                                     )
        self.cache_database_name = cache_database_name
        self.storage_database_name = storage_database_name
        self.job_id = self.generate_job_id()
        self.set_logger()

    def set_logger(self):
        configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})

    def generate_job_id(self):
        return uuid.uuid4().hex

    def setup_database_settings(self, cache_database=None, storage_database=None,
                                ):

        if cache_database and cache_database not in ["mongodb", "elasticsearch"]:
            raise Exception("we only support {} as cache_database".format(",".join(SUPPORTED_DATABASES)))

        if storage_database and storage_database not in ["mongodb", "elasticsearch"]:
            raise Exception("we only support {} as storage_database ".format(",".join(SUPPORTED_DATABASES)))

        if cache_database == "mongodb":
            if self.settings['HTTPCACHE_ENABLED']:
                self.settings['HTTPCACHE_STORAGE'] = MONGODB_DEFAULTS['HTTPCACHE_STORAGE']
        if storage_database == "mongodb":
            self.settings['ITEM_PIPELINES'] = MONGODB_DEFAULTS['ITEM_PIPELINES']

        if cache_database == "elasticsearch":
            if self.settings['HTTPCACHE_ENABLED']:
                self.settings['HTTPCACHE_STORAGE'] = ELASTICSEARCH_DEFAULTS['HTTPCACHE_STORAGE']
        if storage_database == "elasticsearch":
            self.settings['ITEM_PIPELINES'] = ELASTICSEARCH_DEFAULTS['ITEM_PIPELINES']

    def setup_crawler_type_settings(self, crawler_type=None):
        if crawler_type == "websites":
            self.settings.update(WEBSITE_CRAWLER_DEFAULTS)
        elif crawler_type == "feeds":
            self.settings.update(FEEDS_CRAWLER_DEFAULTS)

        if self.settings['HTTPCACHE_ENABLED']:
            if self.cache_database_uri:
                self.settings['INVANA_BOT_SETTINGS']['HTTPCACHE_STORAGE_SETTINGS'][
                    'DATABASE_URI'] = self.cache_database_uri
            if self.cache_database_name:
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS'][
                    'DATABASE_NAME'] = self.cache_database_name

        if self.storage_database_uri:
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['DATABASE_URI'] = self.storage_database_uri

        if self.storage_database_name:
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS'][
                'DATABASE_NAME'] = self.storage_database_name

    def _validate_urls(self, urls):
        if type(urls) is None:
            raise Exception("urls should be list type.")
        if len(urls) is 0:
            raise Exception("urls length should be atleast one.")

    def get_settings(self):
        return self.settings


class InvanaFeedCrawler(InvanaCrawlerBase):

    def crawl_feeds(self, feed_urls=None):
        self.setup_crawler_type_settings(crawler_type="feeds")
        self._validate_urls(feed_urls)

        process = CrawlerProcess(self.settings)
        allowed_domains = []
        for feed_url in feed_urls:
            domain = feed_url.split("://")[1].split("/")[0]  # TODO - clean this
            allowed_domains.append(domain)

        process.crawl(RSSSpider,
                      start_urls=feed_urls,
                      )
        process.start()



from scrapy.crawler import CrawlerRunner
from twisted.internet import reactor


class InvanaJobScheduler(object):
    """
    This will actually schedule the jobs.

    crawler = InvanaWebCrawler(
        cache_database_uri="mongodb://127.0.0.1",
        storage_database_uri="mongodb://127.0.0.1",
        cache_database="mongodb",
        storage_database="mongodb",
    )

    all_jobs = crawler.create_job(
        pipeline=pipeline_data,
        context=context
    )

    scheduler = InvanaJobScheduler(settings=crawler.get_settings(), jobs=crawler.get_jobs())
    scheduler.start()



    """
    jobs = []
    runner = None

    def __init__(self, settings=None):
        self.settings = settings

    def start(self):
        self.start_jobs(jobs=self.jobs)

    def start_jobs(self, jobs=None):
        if self.runner is None:
            self.runner = CrawlerRunner(self.settings)
        for job in jobs:
            spider_cls = job['spider_cls']
            spider_kwargs = job['spider_kwargs']
            self.runner.crawl(spider_cls, **spider_kwargs)
        d = self.runner.join()
        d.addBoth(lambda _: reactor.stop())
        reactor.run()



SUPPORTED_DATABASES = ["mongodb", "elasticsearch"]
SUPPORTED_CRAWLERS = ["website", "feeds"]
MONGODB_DEFAULTS = {
    'ITEM_PIPELINES': {'invana_bot.storages.mongodb.MongoDBPipeline': 1},
    'HTTPCACHE_STORAGE': "invana_bot.httpcache.mongodb.MongoDBCacheStorage",
}

ELASTICSEARCH_DEFAULTS = {
    'ITEM_PIPELINES': {'invana_bot.storages.elasticsearch.ElasticSearchPipeline': 1},
    'HTTPCACHE_STORAGE': "invana_bot.httpcache.elasticsearch.ESCacheStorage",
}

KEYWORD_CRAWLER_DEFAULTS = {

    'INVANA_BOT_SETTINGS': {
        'HTTPCACHE_STORAGE_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_cache_db",
            'COLLECTION_NAME': "web_link",
            "EXPIRY_TIME": 3600
        },
        'ITEM_PIPELINES_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_data",
            'COLLECTION_NAME': "keywords_data"
        }
    }
}

FEEDS_CRAWLER_DEFAULTS = {

    'INVANA_BOT_SETTINGS': {
        'HTTPCACHE_STORAGE_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_cache_db",
            'COLLECTION_NAME': "web_link",
            "EXPIRY_TIME": 3600
        },
        'ITEM_PIPELINES_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_data",
            'COLLECTION_NAME': "feeds_data"
        }
    }
}

WEBSITE_CRAWLER_DEFAULTS = {

    'INVANA_BOT_SETTINGS': {
        'HTTPCACHE_STORAGE_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_cache_db",
            'COLLECTION_NAME': "web_link",
            "EXPIRY_TIME": 3600
        },
        'ITEM_PIPELINES_SETTINGS': {
            'DATABASE_URI': "127.0.0.1",
            'DATABASE_NAME': "crawler_data",
            'COLLECTION_NAME': "website_data"
        }
    }
}



from .base import InvanaCrawlerBase
from invana_bot.pipelines.default import WebCrawlerPipeline
from datetime import datetime


class InvanaWebCrawler(InvanaCrawlerBase):
    """
    Split the bot into website crawler, feeds crawler, api crawler,

    """

    def create_job(self, pipeline=None, context=None):
        self.setup_crawler_type_settings(crawler_type="websites")
        if context is None:
            context = {}
        if 'job_id' not in context.keys():
            context['job_id'] = self.job_id
            context['job_started'] = datetime.now()
        pipeline = WebCrawlerPipeline(pipeline=pipeline, job_id=self.job_id, context=context)
        job = pipeline.run()
        return job





from invana_bot.exceptions import InvalidCrawlerConfig


def validate_config(config=None):
    required_keys = ['start_url', 'data_selectors']
    for key_ in required_keys:
        if key_ not in config.keys():
            raise InvalidCrawlerConfig(
                "Invalid configuration: Required Key {0} not found in the configuration".format(key_))
    # TODO - validate all the data_selectors data aswell
    return True


def process_config(config=None):
    processed_config_dict = {}
    parent_selectors = []
    new_config_selectors = []
    for selector in config.get('data_selectors', []):
        if selector.get('selector_attribute') == 'element':
            parent_selectors.append(selector)

    """
    process the element root selectors (without parent_selector) which might become dictionaries
    """
    if len(parent_selectors) > 0:
        for parent_selector in parent_selectors:
            processed_config_dict[parent_selector.get('id')] = parent_selector
            processed_config_dict[parent_selector.get('id')]['child_selectors'] = []

            for selector in config['data_selectors']:
                if selector.get('parent_selector') == parent_selector.get('id'):
                    processed_config_dict[parent_selector.get('id')]['child_selectors'].append(selector)

    """
    process the elements with no root selectors
    """
    for selector in config.get('data_selectors', []):
        if selector.get('parent_selector') is None and selector.get('selector_attribute') != 'element':
            new_config_selectors.append(selector)

    for k, v in processed_config_dict.items():
        new_config_selectors.append(v)
    config['data_selectors'] = new_config_selectors
    return config



from invana_bot.spiders.default import DefaultPipeletSpider
from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import Rule


class WebCrawlerPipelet(object):
    """

    pipe_data = {  # single pipe

        "pipe_id": "blog-list",
        "start_urls": ["https://blog.scrapinghub.com"],
        "data_extractors": [
            {
                "data_selectors": [
                    {
                        "id": "items",
                        "selector": ".post-listing .post-item",
                        "selector_attribute": "element",
                        "multiple": True
                    },
                    {
                        "id": "url",
                        "selector": ".post-header h2 a",
                        "selector_type": "css",
                        "selector_attribute": "href",
                        "parent_selector": "items",
                        "multiple": False
                    },
                    {
                        "id": "title",
                        "selector": ".post-header h2 a",
                        "selector_type": "css",
                        "selector_attribute": "text",
                        "parent_selector": "items",
                        "multiple": False
                    },
                    {
                        "id": "content",
                        "selector": ".post-content",
                        "selector_type": "css",
                        "selector_attribute": "html",
                        "parent_selector": "items",
                        "multiple": False
                    }
                ],
            }
        ],
        "traversals": [{
            "traversal_type": "pagination",
            "pagination": {
                "selector": ".next-posts-link",
                "selector_type": "css",
                "max_pages": 2
            },
        }]

    }

    context = {
        "client_id": "something"
    }

    """

    def __init__(self, pipe=None, start_urls=None, job_id=None,
                 pipeline=None, context=None):
        """

        :param pipe: single unit of crawling
        :param pipeline: set of units combined to create a flow
        :param context: any extra information user want to send to the crawled data.
        """
        self.pipe = pipe
        self.job_id = job_id
        self.pipeline = pipeline
        self.start_urls = start_urls
        if context:
            self.context = context
        self.validate_pipe()

    def validate_pipe(self):
        must_have_keys = ["pipe_id", "data_extractors"]
        optional_keys = ["traversals"]
        for key in must_have_keys:
            if key not in self.pipe.keys():
                raise Exception(
                    "invalid pipe data, should have the following keys; {}".format(",".join(must_have_keys)))

    def validate_traversal(self):
        pass  # TODO - implement this

    def validate_extractor(self):
        pass  # TODO - implement this

    def get_traversals(self):
        return self.pipe.get("traversals", [])

    def get_extractors(self):
        return self.pipe.get("data_extractors", [])

    def generate_pipe_kwargs(self):
        domains = []

        for url in self.start_urls:
            domain = url.split("://")[1].split("/")[0]  # TODO - clean this
            domains.append(domain)

        extractor = LinkExtractor()
        rules = [
            Rule(extractor, follow=True)  # TODO - add regex types of needed.
        ]

        spider_kwargs = {
            "start_urls": self.start_urls,
            "allowed_domains": domains,
            "rules": rules,
            "pipe": self.pipe,
            "pipeline": self.pipeline,
            "context": self.context
        }
        return spider_kwargs

    def run(self):
        spider_cls = DefaultPipeletSpider
        spider_kwargs = self.generate_pipe_kwargs()
        return {"spider_cls": spider_cls, "spider_kwargs": spider_kwargs}


class WebCrawlerPipeline(object):
    """



    """

    def __init__(self, pipeline=None, job_id=None, context=None):
        self.pipeline = pipeline
        self.job_id = job_id
        self.context = context

    def get_pipelet(self, pipe_id=None):
        pipeline = self.pipeline['pipeline']
        for pipe in pipeline:
            if pipe.get("pipe_id") == pipe_id:
                return pipe
        return

    def run(self):
        pipe = self.pipeline['pipeline'][0]
        invana_pipe = WebCrawlerPipelet(pipe=pipe,
                                        start_urls=self.pipeline['start_urls'],
                                        job_id=self.job_id,
                                        pipeline=self.pipeline,
                                        context=self.context)
        job = invana_pipe.run()
        return job



from scrapy.spiders import XMLFeedSpider, Spider
from invana_bot.utils.url import get_domain
import feedparser


class GenericFeedSpider(XMLFeedSpider):
    name = "generic_feed_spider"
    itertag = 'item'
    iterator = 'xml'  # This is actually unnecessary, since it's the default value

    def get_or_none(self, selector):
        try:
            return selector.extract_first()
        except Exception as e:
            print(e)
            return None

    def parse_node(self, response, node):
        title = self.get_or_none(node.select('title/text()'))

        url = self.get_or_none(node.select('link/text()'))
        description = self.get_or_none(node.select('description/text()'))
        pub_date = self.get_or_none(node.select('pubDate/text()'))
        category = self.get_or_none(node.select('category/text()'))
        # image = node.select('item/media:content/url')

        item = {}
        item['title'] = title
        item['url'] = url

        item['pub_date'] = pub_date
        item['category'] = category
        item['description'] = description
        item['domain'] = get_domain(response.url)
        return item


class RSSSpider(Spider):
    """

    https://gist.github.com/nyov/9c70e780ea80204559d6da5525228702

    """
    name = "rss"

    def parse_feed(self, feed):
        """ Parse RSS/Atom feed using feedparser
        """
        data = feedparser.parse(feed)
        if data.bozo:
            if (hasattr(data.bozo_exception, 'getLineNumber') and
                    hasattr(data.bozo_exception, 'getMessage')):
                line = data.bozo_exception.getLineNumber()
                # segment = feed.split('\n')[line - 1]
            # could still try to return data. not necessarily completely broken
            return None
        return data

    def parse(self, response):
        # parse downloaded content with feedparser (NOT re-downloading with feedparser)
        feed = self.parse_feed(response.body)
        if feed:
            # grab some feed elements
            # - https://pythonhosted.org/feedparser/common-rss-elements.html
            # - https://pythonhosted.org/feedparser/common-atom-elements.html

            # ns = feed.namespaces
            # feed_title = feed.feed.title
            # feed_link = feed.feed.link
            # feed_desc = feed.feed.description

            for entry in feed.entries:
                # have content?
                content = entry.get('content')
                if content:
                    # content = content[0]
                    content = content[0]['value']

                item = {
                    # global feed data
                    # 'feed_title': feed_title,
                    # 'feed_link': feed_link,
                    # 'feed_description': feed_desc,
                    #
                    # item entry data
                    # 'url': response.url,
                    'url': entry.link,
                    'title': entry.title,
                    'domain': get_domain(response.url),
                    'description': entry.description,
                    # 'date': entry.published,
                    # 'date': entry.published_parsed,
                    'pub_date': entry.updated_parsed,

                    # optional
                    'content': content,
                    'type': entry.get('dc_type'),
                }

                yield item



#!/usr/bin/env python

from setuptools import setup

setup(name='invana-bot',
      version='0.0.0a',
      description='A batteries included crawler framework built on top of scrapy for scale and intelligent crawling.',
      author='Ravi Raja Merugu',
      author_email='rrmerugu@gmail.com',
      url='https://github.com/invanalabs/invana-bot',
      packages=[
          'tests',
          'invana_bot',
          'invana_bot.extractors',
          'invana_bot.fields',
          'invana_bot.httpcache',
          'invana_bot.pipelines',
          'invana_bot.spiders',
          'invana_bot.storages',
          'invana_bot.utils',
          'invana_bot.crawlers',
          'invana_bot.schedulers',
      ],
      install_requires=['Scrapy==1.6.0',
                        'pymongo',
                        "feedparser==5.2.1"
                        ]
      )
